"""pytest for SCConfigManager class."""
import datetime as dt
import sys

from sc_utility import DateHelper, SCConfigManager

# Remove the period if running this in the debugger
from .config_schemas import ConfigSchema

CONFIG_FILE = "tests/config.yaml"


print("Running test for SCConfigManager...")
# Get our default schema, validation schema, and placeholders
schemas = ConfigSchema()

# Initialize the SC_ConfigManager class
try:
    config = SCConfigManager(
        config_file=CONFIG_FILE,
        default_config=schemas.default,
        validation_schema=schemas.validation,
        placeholders=schemas.placeholders
    )
except RuntimeError as e:
    print(f"Configuration file error: {e}", file=sys.stderr)
    sys.exit(1)


def test_get_value():
    """Test reading configuration values from the config file."""
    value1 = config.get("Testing", "Value1")
    value2 = config.get("Testing", "Value2")

    string1 = config.get("Testing", "String1")
    string2 = config.get("Testing", "String2")

    assert value1 == value2, "Value1 and Value2 should be equal"
    assert string1 == string2, "String1 and String2 should be equal"

    print("Configuration values read successfully:")


def test_load_config():
    """Test loading the configuration file."""
    assert config.load_config(), "Failed to load configuration"

    print("Configuration loaded successfully.")


def test_check_for_config_changes():
    """Test checking for changes in the configuration file."""
    # Create a fake last check time well in the past
    last_check = DateHelper.now() - dt.timedelta(days=365)
    assert config.check_for_config_changes(last_check) is not None, "Configuration changes were not detected"

    print("Configuration changes detected successfully.")


def test_register_logger():
    """Test registering a logger with the configuration manager will be tested in test_sc_logging."""
    # Nothing to do here.


def test_check_for_placeholders():
    """Test checking for placeholders in the configuration."""
    assert not config.check_for_placeholders(schemas.placeholders), "Placeholders were found in the configuration"

    print("No placeholders found in the configuration.")


def test_get_logger_settings():
    """Test getting logger settings from the configuration."""
    logger_settings = config.get_logger_settings()

    assert isinstance(logger_settings, dict), "Logger settings should be a dictionary"
    assert "logfile_name" in logger_settings, "Logger settings should contain 'logfile_name'"

    print("Logger settings retrieved successfully:", logger_settings)



"""Some basic shorts for handling dates."""

import datetime as dt
from pathlib import Path


class DateHelper:
    """
    Class for simplyify date operations.

    This class provides static methods to handle date formatting, parsing, and calculations. It defaults to the "YYYY-MM-DD" format for date strings, but this can be overridden by passing a different format string.
    It also handles timezone-aware dates by using the local timezone of the system when parsing and formatting dates.
    """

    @staticmethod
    def format_date(date_obj: dt.date | dt.datetime, date_format: str = "%Y-%m-%d") -> str | None:
        """
        Format a date object to a string.

        Args:
            date_obj (date | datetime): The date or datetime object to format.
            date_format (str, optional): The format string to use for formatting the date or datetime.

        Returns:
            date_str (str): The formatted date string, or None if date_obj is None.
        """
        if date_obj is None:
            return None
        return date_obj.strftime(date_format)

    @staticmethod
    def parse_date(date_str: str, date_format: str = "%Y-%m-%d") -> dt.date | dt.datetime | None:
        """
        Parse a date string to a date or datetime object.

        Args:
            date_str (str): The date string to parse.
            date_format (Optional[str], optional): The format string to use for parsing the date. Defaults to "%Y-%m-%d".

        Returns:
            date_obj (date | datetime): A date or datetime object representing the parsed date_str, or None if date_str is empty.
        """
        local_tz = dt.datetime.now().astimezone().tzinfo
        if not date_str:
            return None
        parsed_dt = dt.datetime.strptime(date_str, date_format).replace(tzinfo=local_tz)

        # If the date_format string conatins only date components (like "%Y-%m-%d"), return a date object.
        # If it contains time components (like "%Y-%m-%d %H:%M:%S"), return a datetime object.
        if "%H" in date_format or "%M" in date_format or "%S" in date_format:
            return parsed_dt
        return parsed_dt.date()

    @staticmethod
    def days_between(start_date: dt.date | dt.datetime, end_date: dt.date | dt.datetime) -> int | None:
        """
        Calculate the number of days between two date or datetime objects.

        Args:
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            difference (int): The number of days between the two dates, or None if either date is None.
        """
        if start_date is None or end_date is None:
            return None
        if isinstance(start_date, dt.datetime):
            start_date = start_date.date()
        if isinstance(end_date, dt.datetime):
            end_date = end_date.date()
        return (end_date - start_date).days

    @staticmethod
    def add_days(start_date: dt.date | dt.datetime, days: int) -> dt.date | dt.datetime | None:
        """
        Add days to a date or datetime object.

        Args:
            start_date (date | datetime): The date/datetime to which days will be added.
            days (int): The number of days to add.

        Returns:
            result (date | datetime) : A new date or datetime object with the added days, or None if start_date or days is None.

        """
        if start_date is None or days is None:
            return None
        return start_date + dt.timedelta(days=days)

    @staticmethod
    def is_valid_date(date_str: str, date_format: str = "%Y-%m-%d") -> bool:
        """
        Check if a date or datetime string is valid according to the specified format.

        Args:
            date_str (str): The date string to check.
            date_format (Optional[str], optional): The format string to use for checking the date. Defaults to "%Y-%m-%d".

        Returns:
            result (bool): True if the date string is valid, False otherwise.
        """
        local_tz = dt.datetime.now().astimezone().tzinfo
        try:
            dt.datetime.strptime(date_str, date_format).replace(tzinfo=local_tz)
        except ValueError:
            return False
        else:
            return True

    @staticmethod
    def today() -> dt.date:
        """
        Get today's date.

        Returns:
            result (date): Today's date as a date object, using the local timezone.
        """
        local_tz = dt.datetime.now().astimezone().tzinfo
        return dt.datetime.now(tz=local_tz).date()

    @staticmethod
    def now() -> dt.datetime:
        """
        Get today's date and time.

        Returns:
            result (datetime): Today's date and time as a date object, using the local timezone.
        """
        local_tz = dt.datetime.now().astimezone().tzinfo
        return dt.datetime.now(tz=local_tz)

    @staticmethod
    def today_add_days(days: int) -> dt.date:
        """
        Get today's date ofset by days.

        Args:
            days (int): The number of days to offset from today. Can be positive or negative

        Returns:
            result (date): Today's date offset by the specified number of days.
        """
        date_today = DateHelper.today()
        return DateHelper.add_days(date_today, days)  # type: ignore[call-arg]

    @staticmethod
    def today_str(date_format: str | None = "%Y-%m-%d") -> str:
        """
        Get today's date in string format.

        Args:
            date_format (Optional[str], optional): The format string to use for formatting the date.

        Returns:
            result (str): Today's date as a formatted string, using the specified date format.
        """
        date_today = DateHelper.today()
        return DateHelper.format_date(date_today, date_format)  # type: ignore[call-arg]

    @staticmethod
    def now_str(datetime_format: str | None = "%Y-%m-%d %H:%M:%S") -> str:
        """
        Get the current time in string format.

        Args:
            datetime_format (Optional[str], optional): The format string to use for formatting the datetime.

        Returns:
            result (str): Current time as a formatted string, using the specified datetime format.
        """
        time_now = DateHelper.now()
        return DateHelper.format_date(time_now, datetime_format)  # type: ignore[call-arg]

    @staticmethod
    def get_file_date(file_path: str | Path) -> dt.date | None:
        """
        Get the last modified date of a file.

        Args:
            file_path (str | Path): Path to the file. Can be a string or a Path object.

        Returns:
            date_obj (date): The last modified date of the file as a date object, or None if the file does not exist.
        """
        if isinstance(file_path, str):
            file_path = Path(file_path)

        if not file_path.exists():
            return None

        local_tz = dt.datetime.now().astimezone().tzinfo
        return dt.datetime.fromtimestamp(file_path.stat().st_mtime, tz=local_tz).date()

    @staticmethod
    def get_file_datetime(file_path: str | Path) -> dt.datetime | None:
        """
        Get the last modified datetime of a file.

        Args:
            file_path (str | Path): Path to the file. Can be a string or a Path object.

        Returns:
            datetime_obj (datetime): The last modified datetime of the file as a date object, or None if the file does not exist.
        """
        if isinstance(file_path, str):
            file_path = Path(file_path)

        if not file_path.exists():
            return None

        local_tz = dt.datetime.now().astimezone().tzinfo
        return dt.datetime.fromtimestamp(file_path.stat().st_mtime, tz=local_tz)



import datetime as dt
import sys

from sc_utility import CSVReader

CSV_FILE = "tests/test_csv.csv"
NEW_CSV_FILE = "tests/test_csv_new.csv"

header_config = [
    {
        "name": "Symbol",
        "type": "str",
        "match": True,
        "sort": 2,
    },
    {
        "name": "Date",
        "type": "date",
        "format": "%Y-%m-%d",
        "match": True,
        "sort": 1,
        "minimum": dt.date(2025, 7, 8),
    },
    {
        "name": "Name",
        "type": "str",
    },
    {
        "name": "Currency",
        "type": "str",
    },
    {
        "name": "Price",
        "type": "float",
        "format": ".2f",
    },
]

extra_data = [
    {
        "Symbol": "AAPL",
        "Date": dt.date(2023, 10, 1),
        "Name": "Apple Inc.",
        "Currency": "USD",
        "Price": 150.00,
    },
    {
        "Symbol": "AAPL",
        "Date": dt.date(2023, 11, 1),
        "Name": "Apple Inc.",
        "Currency": "USD",
        "Price": 155.00,
    },
]


def sum_list_values(data: list[dict]) -> float:
    """
    Helper function to sum values in a list of dict.

    Args:
        data (dict): The data to sum values from.

    Returns:
        float: The sum of all the Price keys in the list.
    """
    sum_value = 0
    for row in data:
        sum_value += row.get("Price", 0.0)
    return sum_value


def test_read_csv():
    """Test the CSVReader class reading from a CSV file."""
    # Create an instance of the CSVReader class
    try:
        csv_reader = CSVReader(CSV_FILE, header_config)
        csv_data = csv_reader.read_csv()
    except (ImportError, TypeError, ValueError) as e:
        print(f"Error reading file: {e}")
        sys.exit(1)
    else:
        assert isinstance(csv_data, list), "CSV data should be a list"
        assert len(csv_data) > 0, "CSV data should not be empty"
        assert sum_list_values(csv_data) > 147, "Sum of values in the sheet should be greater than 147"


def test_sort_csv_data():
    try:
        csv_reader = CSVReader(CSV_FILE, header_config)
        csv_data = csv_reader.read_csv()
        if not csv_data:
            print("CSV data is empty, cannot sort data.")
            sys.exit(1)
        csv_data = csv_reader.sort_csv_data(csv_data)
    except (ImportError, TypeError, ValueError) as e:
        print(f"Error reading file: {e}")
        sys.exit(1)
    else:
        assert isinstance(csv_data, list), "CSV data should be a list"
        assert len(csv_data) > 0, "CSV data should not be empty"
        sum_value = round(sum_list_values(csv_data), 0)
        assert sum_value == 771, "Sum of values in the sheet should be 771"


def test_merge_data_sets():
    try:
        csv_reader = CSVReader(CSV_FILE, header_config)
        csv_data = csv_reader.read_csv()
        if not csv_data:
            print("CSV data is empty, cannot merge with extra data.")
            sys.exit(1)
        csv_data = csv_reader.merge_data_sets(csv_data, extra_data)
    except (ImportError, TypeError, ValueError) as e:
        print(f"Error reading file: {e}")
        sys.exit(1)
    else:
        assert isinstance(csv_data, list), "CSV data should be a list"
        assert len(csv_data) > 0, "CSV data should not be empty"
        sum_value = round(sum_list_values(csv_data), 0)
        assert sum_value == 1076, "Sum of values in the sheet should be 1076"


def test_trim_csv_data():
    try:
        csv_reader = CSVReader(CSV_FILE, header_config)
        csv_data = csv_reader.read_csv()
        if not csv_data:
            print("CSV data is empty, cannot trim data.")
            sys.exit(1)
        csv_data = csv_reader.trim_csv_data(csv_data)
    except (ImportError, TypeError, ValueError) as e:
        print(f"Error reading file: {e}")
        sys.exit(1)
    else:
        assert isinstance(csv_data, list), "CSV data should be a list"
        assert len(csv_data) == 3, "CSV data should have 3 rows after trimming"
        sum_value = round(sum_list_values(csv_data), 0)
        assert sum_value == 145, "Sum of values in the sheet should be 145"


def test_write_csv():
    try:
        csv_reader = CSVReader(CSV_FILE, header_config)
        csv_data = csv_reader.read_csv()
        if not csv_data:
            print("CSV data is empty, cannot write data.")
            sys.exit(1)
        csv_data = csv_reader.merge_data_sets(csv_data, extra_data)
        csv_data = csv_reader.trim_csv_data(csv_data)
        result = csv_reader.write_csv(csv_data, NEW_CSV_FILE)
    except (ImportError, TypeError, ValueError) as e:
        print(f"Error reading file: {e}")
        sys.exit(1)
    else:
        assert result, "write_csv() shoudl return True"



"""pytest for DateHelper class."""
import datetime as dt
from pathlib import Path

from sc_utility import DateHelper

CONFIG_FILE = "tests/config.yaml"


def test_format_date():
    """Test format_date."""
    date_obj = dt.date(2025, 2, 1)
    formatted_date = DateHelper.format_date(date_obj)
    assert formatted_date == "2025-02-01", "Formatted date should be '2025-02-01'"


def test_parse_date():
    """Test parse_date."""
    date_str = "2025-02-04"
    parsed_date = DateHelper.parse_date(date_str)
    assert parsed_date == dt.date(2025, 2, 4), "Parsed date should be 2025-02-04"


def test_days_between():
    """Test calculating the number of days between two dates."""
    d1 = dt.date(2024, 1, 1)
    d2 = dt.date(2024, 1, 10)
    assert DateHelper.days_between(d1, d2) == 9, "Days between 2024-01-01 and 2024-01-10 should be 9"


def test_add_days():
    """Test adding days to a date."""
    start = dt.date(2024, 1, 1)
    result = DateHelper.add_days(start, 5)
    assert result == dt.date(2024, 1, 6), "Adding 5 days to 2024-01-01 should result in 2024-01-06"


def test_is_valid_date():
    """Test is_valid_date."""
    date_str = "2025-05-04"
    datetime_str = "2025-05-04 12:10:08"
    assert DateHelper.is_valid_date(date_str, "%Y-%m-%d"), "Date '2025-05-04' should be valid"
    assert DateHelper.is_valid_date(datetime_str, "%Y-%m-%d %H:%M:%S"), "Datetime '2025-05-04  12:10:08' should be valid"


def test_today():
    """Test today()."""
    local_tz = dt.datetime.now().astimezone().tzinfo
    today_date = dt.datetime.now(tz=local_tz).date()
    assert DateHelper.today() == today_date, "Today should return the current date"


def test_now():
    """Test today()."""
    local_tz = dt.datetime.now().astimezone().tzinfo
    time_now = dt.datetime.now(tz=local_tz)
    time_now_str = DateHelper.format_date(time_now, "%Y-%m-%d %H:%M")
    func_time_now_str = DateHelper.format_date(DateHelper.now(), "%Y-%m-%d %H:%M")
    assert time_now_str == func_time_now_str, "Now should return the current time"


def test_today_add_days():
    """Test today() with add_days."""
    today_date = DateHelper.today()
    future_date = DateHelper.add_days(today_date, 5)
    assert future_date == today_date + dt.timedelta(days=5), "Adding 5 days to today should return the correct date"


def test_today_str():
    """Test getting string representation of today's date."""
    today_str = DateHelper.today_str()
    formatted_today = DateHelper.format_date(DateHelper.today())
    assert today_str == formatted_today, "Today's string representation should match the formatted date"


def test_now_str():
    """Test getting string representation of the current time."""
    local_tz = dt.datetime.now().astimezone().tzinfo
    time_now = dt.datetime.now(tz=local_tz)
    time_str = DateHelper.format_date(time_now, "%Y-%m-%d %H:%M")
    formatted_now = DateHelper.format_date(DateHelper.now(), "%Y-%m-%d %H:%M")
    assert time_str == formatted_now, "Current time string representation should match the formatted datetime"


def test_get_file_date():
    """Test getting the file date."""
    file_path = Path(CONFIG_FILE)
    file_date = DateHelper.get_file_date(CONFIG_FILE)

    local_tz = dt.datetime.now().astimezone().tzinfo
    check_file_date = dt.datetime.fromtimestamp(file_path.stat().st_mtime, tz=local_tz).date()
    assert file_date == check_file_date, "File date should match the last modified date of the file"


def test_get_file_datetime():
    """Test getting the file datetime."""
    file_path = Path(CONFIG_FILE)
    file_date = DateHelper.get_file_datetime(CONFIG_FILE)

    local_tz = dt.datetime.now().astimezone().tzinfo
    check_file_datetime = dt.datetime.fromtimestamp(file_path.stat().st_mtime, tz=local_tz)
    assert file_date == check_file_datetime, "File datetime should match the last modified datetime of the file"



"""pytest for ShellyControl class."""
import sys

from sc_utility import SCConfigManager, SCLogger, ShellyControl

# Remove the period if running this in the debugger
from .config_schemas import ConfigSchema

CONFIG_FILE = "tests/config.yaml"
DEVICE_CLIENTNAME = "Device Test 1"

print("Running test for ShellyCOntrol...")

# Get our default schema, validation schema, and placeholders
schemas = ConfigSchema()

# Initialize the SC_ConfigManager class
try:
    config = SCConfigManager(
        config_file=CONFIG_FILE,
        default_config=schemas.default,
        validation_schema=schemas.validation,
        placeholders=schemas.placeholders
    )
except RuntimeError as e:
    print(f"Configuration file error: {e}", file=sys.stderr)
    sys.exit(1)
else:
    assert config is not None, "ConfigManager should be initialized"


# Initialize the SC_Logger class
try:
    logger = SCLogger(config.get_logger_settings())
except RuntimeError as e:
    print(f"Logger initialisation error: {e}", file=sys.stderr)
    sys.exit(1)
else:
    assert logger is not None, "Logger should be initialized"

# Create the ShellyControl object
shelly_settings = config.get_shelly_settings()
assert shelly_settings is not None, "Shelly settings should be initialized"

# Initialize the SC_ShellyControl class
try:
    shelly_control = ShellyControl(logger, shelly_settings)  # type: ignore[call-arg]
except RuntimeError as e:
    print(f"Shelly control initialization error: {e}", file=sys.stderr)
    sys.exit(1)
else:
    assert shelly_control is not None, "ShellyControl should be initialized"


def test_get_device():
    """Test function for Shelly control."""
    assert shelly_control is not None, "ShellyControl should be initialized"
    try:
        device = shelly_control.get_device(DEVICE_CLIENTNAME)
    except RuntimeError as e:
        print(e, file=sys.stderr)
        sys.exit(1)
    else:
        assert device is not None, f"Device {DEVICE_CLIENTNAME} should be found"
        assert isinstance(device, dict), "Device should be a dictionary"
        assert device.get("Index") == 0, "Device Index should be 0"
        assert device.get("ClientName") == DEVICE_CLIENTNAME, f"Device identity should be {DEVICE_CLIENTNAME}"
        assert device.get("Model") == "Shelly2PMG3", "Device Model should be Shelly2PMG3"
        assert device.get("Inputs") == 2, "Device input count should be 2"
        assert device.get("Outputs") == 2, "Device output count should be 2"
        assert device.get("Meters") == 2, "Device meter  count should be 2"


def test_get_device_information():
    """Test function for Shelly control."""
    assert shelly_control is not None, "ShellyControl should be initialized"
    try:
        device = shelly_control.get_device(DEVICE_CLIENTNAME)
        device_info = shelly_control.get_device_information(device)
    except RuntimeError as e:
        print(e, file=sys.stderr)
        sys.exit(1)
    else:
        assert device_info is not None, f"Device {DEVICE_CLIENTNAME} information should be found"
        assert isinstance(device_info, dict), f"Device {DEVICE_CLIENTNAME} information should be a dict"
        assert device_info.get("Model") == "Shelly2PMG3", "Device information should contain model information"
        assert len(device_info.get("Inputs")) == 2, "Device information should contain 2 inputs"  # type: ignore[call-arg]
        assert len(device_info.get("Outputs")) == 2, "Device information should contain 2 outputs"  # type: ignore[call-arg]
        assert len(device_info.get("Meters")) == 2, "Device information should contain 2 meters"  # type: ignore[call-arg]


def test_get_device_status():
    """Test function for Shelly control."""
    assert shelly_control is not None, "ShellyControl should be initialized"
    try:
        device = shelly_control.get_device(DEVICE_CLIENTNAME)
        result = shelly_control.get_device_status(device)
    except RuntimeError as e:
        print(e, file=sys.stderr)
        sys.exit(1)
    else:
        assert result is not None, f"Device {DEVICE_CLIENTNAME} status should be found"

    device_input = shelly_control.get_device_component("input", "Device 1.Input 1")
    assert device_input is not None, "Device input should be found"
    assert device_input.get("State") is not None, "Device input state should be found"
    assert isinstance(device_input.get("State"), bool), "Device input state should be a boolean"

    device_output = shelly_control.get_device_component("output", "Device 1.Output 2")
    assert device_output is not None, "Device output should be found"
    assert device_output.get("State") is not None, "Device output state should be found"
    assert isinstance(device_output.get("State"), bool), "Device output state should be a boolean"

    device_meter = shelly_control.get_device_component("meter", "Meter 1")  # Auto generated name
    assert device_meter is not None, "Device meter should be found"


def test_refresh_all_device_statuses():
    """Test function for refreshing all device statuses."""
    assert shelly_control is not None, "ShellyControl should be initialized"
    try:
        shelly_control.refresh_all_device_statuses()
    except RuntimeError as e:
        print(e, file=sys.stderr)
        sys.exit(1)


def test_is_device_online():
    """Test function for checking if a device is online."""
    assert shelly_control is not None, "ShellyControl should be initialized"
    try:
        is_online = shelly_control.is_device_online(DEVICE_CLIENTNAME)
    except RuntimeError as e:
        print(e, file=sys.stderr)
        sys.exit(1)
    else:
        assert is_online, f"Device {DEVICE_CLIENTNAME} should be online"


def test_print_device_status():
    """Test function for printing device status."""
    assert shelly_control is not None, "ShellyControl should be initialized"
    try:
        device_status = shelly_control.print_device_status(DEVICE_CLIENTNAME)
    except RuntimeError as e:
        print(e, file=sys.stderr)
        sys.exit(1)
    else:
        assert device_status is not None, f"Device {DEVICE_CLIENTNAME} status should be printed"
        assert isinstance(device_status, str), f"Device {DEVICE_CLIENTNAME} status should be a string"
        assert device_status.find("Generation: ") != -1, "Device status should contain generation information"


def test_print_model_library():
    """Test function for printing model library."""
    assert shelly_control is not None, "ShellyControl should be initialized"
    try:
        model_library = shelly_control.print_model_library(mode_str="brief")
    except RuntimeError as e:
        print(e, file=sys.stderr)
        sys.exit(1)
    else:
        assert model_library is not None, "Model library should be printed"
        assert isinstance(model_library, str), "Model library should be a string"
        assert model_library.find("Model: Shelly2PMG3") != -1, "Model library should contain Shelly2PMG3 model"


def test_change_output():
    """Test function for changing output state."""
    assert shelly_control is not None, "ShellyControl should be initialized"
    try:
        output_identity = "Device 1.Output 1"  # Example output identity
        output_obj = shelly_control.get_device_component("output", output_identity)
        current_state = output_obj["State"]
        result, did_change = shelly_control.change_output(output_identity, not current_state)
    except RuntimeError as e:
        print(e, file=sys.stderr)
        sys.exit(1)
    else:
        assert result, f"Output {output_identity} state changed successfully"
        assert did_change, f"Output {output_identity} state should have changed"
        # Optionally, you can check the new state of the output
        new_state = shelly_control.get_device_component("output", output_identity)["State"]
        assert new_state != current_state, "Output state should be changed"


test_get_device()
test_get_device_information()
test_get_device_status()
test_is_device_online()
test_print_model_library()
test_change_output()



from enum import StrEnum
from pathlib import Path

from sc_utility import DateHelper, JSONEncoder

JSON_FILE = "tests/test_json.json"


class APIMode(StrEnum):
    LIVE = "Live Prices"
    OFFLINE = "Offline Cache"
    DISABLED = "Pricing Disabled"


data_obj = {
    "string": "Hello, World!",
    "integer": 42,
    "float": 123.456,
    "datetime": DateHelper.now(),
    "date": DateHelper.today(),
    "api_mode": APIMode.LIVE,
}


def test_write_json():
    """Test the JSONEncoder class writing to and reading from a JSON file."""
    # Save the to a JSON file
    file_path = Path(JSON_FILE)
    result = JSONEncoder.save_to_file(data_obj, file_path)
    assert result is True, "Failed to save data to JSON file"


def test_read_json():
    """Test the JSONEncoder class reading from a JSON file."""
    file_path = Path(JSON_FILE)
    loaded_data = JSONEncoder.read_from_file(file_path)
    assert isinstance(loaded_data, dict), "Loaded data should be a dictionary"
    assert loaded_data.get("string") == data_obj.get("string"), "String value mismatch"
    assert loaded_data.get("integer") == data_obj.get("integer"), "Integer value mismatch"
    assert loaded_data.get("float") == data_obj.get("float"), "Float value mismatch"
    assert loaded_data.get("datetime") == data_obj.get("datetime"), "Datetime value mismatch"
    assert loaded_data.get("date") == data_obj.get("date"), "Date value mismatch"
    assert loaded_data.get("api_mode") == data_obj.get("api_mode"), "API mode value mismatch"




import threading
from http.server import BaseHTTPRequestHandler
from urllib.parse import parse_qs

from sc_utility.sc_logging import SCLogger

DEFAULT_WEBHOOK_PATH = "/shelly/webhook"


class _ShellyWebhookHandler(BaseHTTPRequestHandler):
    # `server` attribute will be a ThreadingHTTPServer with `.controller` and `.config_path` attributes attached.

    @property
    def app_wake_event(self) -> threading.Event:
        return getattr(self.server, "app_wake_event", None)  # pyright: ignore[reportReturnType]

    @property
    def logger(self) -> SCLogger:
        return getattr(self.server, "logger", None)  # pyright: ignore[reportReturnType]

    def _ok(self, body: bytes = b"OK"):
        self.send_response(200)
        self.send_header("Content-Type", "text/plain; charset=utf-8")
        self.send_header("Content-Length", str(len(body)))
        self.end_headers()
        self.wfile.write(body)

    def do_GET(self):
        # Validate path like do_POST
        try:
            expected_path = getattr(self.server, "webhook_path", DEFAULT_WEBHOOK_PATH)
        except AttributeError:
            expected_path = DEFAULT_WEBHOOK_PATH

        # Split path and query
        path_only = self.path.split("?")[0]
        query_string = self.path[len(path_only):].lstrip("?") if "?" in self.path else ""

        if expected_path and path_only != expected_path:
            self.send_error(404, "Not Found")
            return

        args = {}
        if query_string:
            args = parse_qs(query_string)
            self.logger.log_message(f"Received webhook GET request {self.path}. Arguments: {args}.", "debug")
        else:
            self.logger.log_message(f"Received webhook GET request {self.path} with no arguments.", "debug")

        # Report the event and wake the controller's run loop immediately
        try:
            controller = getattr(self.server, "controller", None)
            if controller is not None:
                # Add self.path as the first key/value pair to args
                args["path"] = self.path  # pyright: ignore[reportArgumentType]
                controller._push_webhook_event(args)  # noqa: SLF001

            wake_event = getattr(self.server, "app_wake_event", None)
            if wake_event is not None:
                wake_event.set()
        except AttributeError:
            pass

        self._ok(b"LightingController webhook up")



"""Common utility functions and classes used by other classes in the sc_utility package."""

import ipaddress
import os
import platform
import re
import subprocess  # noqa: S404
from pathlib import Path

import httpx
import validators


class SCCommon:
    """Common utility functions and classes used by other classes in the sc_utility package."""

    @staticmethod
    def is_valid_hostname(target: str) -> bool:
        """Return whether target is a valid IPv4, IPv6, or DNS hostname.

        Args:
            target: The target string to validate.

        Returns:
            A boolean indicating validity.
        """
        result, _ = SCCommon.check_hostname_and_type(target)
        return result

    @staticmethod
    def check_hostname_and_type(target: str) -> tuple[bool, str | None]:
        """Return whether target is a valid IPv4, IPv6, or DNS hostname. Also returns the type.

        Args:
            target: The target string to validate.

        Returns:
            A tuple containing a boolean indicating validity and a string indicating the type ('ipv4', 'ipv6', or 'hostname').
        """
        # Make sure the target is a string
        if not isinstance(target, str):
            return False, None

        # Check strict IPv4
        try:
            ipaddress.IPv4Address(target)
        except ValueError:
            pass
        else:
            if target.count(".") == 3:
                return True, "ipv4"

        # Check strict IPv6
        try:
            ipaddress.IPv6Address(target)
        except ValueError:
            pass
        else:
            # If it is a valid IPv6 address, return True
            return True, "ipv6"

        # Reject if it looks like a malformed IP (like 192.168.1 or 256.1.1.1)
        if re.fullmatch(r"[0-9.]+", target):
            return False, None

        # Validate hostname using validators library
        if validators.domain(target) or validators.hostname(target, rfc_1034=True):
            return True, "hostname"

        return False, None

    @staticmethod
    def ping_host(ip_address: str, timeout: int = 1) -> bool:
        """Pings an IP address and returns True if the host is responding, False otherwise.

        Args:
            ip_address: The IP address to ping.
            timeout: Timeout in seconds for the ping response. Default is 1 second.

        Raises:
            RuntimeError: If the IP address is invalid or the ping system call fails.

        Returns:
            result (bool): True if the host responds, False otherwise.
        """
        # Determine the ping command based on the operating system
        param = "-n" if platform.system().lower() == "windows" else "-c"

        if not SCCommon.is_valid_hostname(ip_address):
            error_msg = f"Invalid IP address: {ip_address}"
            raise RuntimeError(error_msg)

        command = ["ping", param, "1", "-W", str(timeout), ip_address]

        try:
            # Run the ping command using subprocess for better security
            result = subprocess.run(command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, shell=False, check=False)  # noqa: S603
            response_code = result.returncode
        except OSError as e:
            error_msg = f"Error pinging {ip_address}: {e}"
            raise RuntimeError(error_msg) from e
        else:
            # Return True if the ping was successful (exit code 0)
            return response_code == 0

    @staticmethod
    def check_internet_connection(urls=None, timeout: int = 3) -> bool:
        """Check if the system has an active internet connection by trying to open a connection to common websites.

        Args:
            urls (list): A list of URLs to check for internet connectivity. Defaults to common DNS servers and websites.
            timeout (int): The timeout in seconds for each request.

        Returns:
            True if the system is connected to the internet, False otherwise.
        """
        if urls is None:
            urls = [
                "https://1.1.1.1",         # Cloudflare DNS
                "https://8.8.8.8",         # Google DNS
                "https://www.google.com",
                "https://www.cloudflare.com"
            ]

        for url in urls:
            try:
                response = httpx.get(url, timeout=timeout, follow_redirects=True)
                if response.status_code < 400:
                    return True
            except httpx.RequestError:
                continue
        return False

    @staticmethod
    def get_os() -> str:
        """Return the name of the operating system.

        Returns:
            The name of the operating system in lowercase.
        """
        # Get the platform name and convert it to lowercase
        platform_name = platform.system().lower()

        if platform_name == "darwin":
            platform_name = "macos"

        return platform_name

    @staticmethod
    def is_probable_path(possible_path: str | Path) -> bool:
        """Check if the given string or Path object is likely to be a file path.

        This method checks if the string is an absolute path, contains a path separator, or has a file extension.

        Args:
            possible_path: The string to check.

        Returns:
            True if the string is likely a file path, False otherwise.
        """
        max_path = 260 if SCCommon.get_os() == "windows" else os.pathconf("/", "PC_PATH_MAX")

        path_obj = None
        if isinstance(possible_path, Path):
            path_str = str(possible_path)
            path_obj = possible_path
        else:
            path_str = possible_path

        if len(path_str) > max_path:
            # If the path is longer than the maximum allowed path length, it cannot be a valid path
            return False

        if path_obj is None:
            path_obj = Path(possible_path)

        # Check if it's absolute, or contains a path separator, or has a file extension
        if path_obj.is_absolute():
            return True

        if "/" in path_str or "\\" in path_str:
            return True

        # Check if the path has a file extension
        return bool(path_obj.suffix and path_obj.suffix.lower() is not None)

    @staticmethod
    def get_project_root(marker_files=("pyproject.toml", ".project_root", "uv.lock", ".git")) -> Path:
        """Return the root folder of the Python project.

        Args:
            marker_files (tuple): A tuple of file names that indicate the project root.

        Raises:
            RuntimeError: If the project root cannot be found.

        Returns:
            root_dir (Path): The root folder of the Python project as a Path object.
        """
        path = Path(__file__).resolve()

        # Walk upwards until we find a marker file
        for parent in [path, *list(path.parents)]:
            for marker in marker_files:
                if (parent / marker).exists():
                    return parent

        error_msg = f"Project root not found. Looked for markers: {marker_files}"
        raise RuntimeError(error_msg)

    @staticmethod
    def select_file_location(file_name: str) -> Path | None:
        """Select the file location for the given file name. It resolves an absolute path for the file_name as follows.

        1. If file_name is an absolute path, return it as a Path object.
        2. If file_name is a relative path (contains parent directories), return the absolute path based on the current working directory.
        3. If file_name is just a file name, look for it in the current working directory first, then in the root directory.

        The root directly is defined as the directory containing the main script being executed (the module containing __main__).

        Raises:
            RuntimeError: If the project root cannot be determined.

        Args:
            file_name: The name of the file to locate. Can be just a file name, or a relative or absolute path.

        Returns:
            file_path (Path): The full path to the file as a Path object. None if the file_name does not appear to be a path.
        """
        # Look at the file_name and see if it looks like a path
        if not SCCommon.is_probable_path(file_name):
            return None

        # Check to see if file_name is a full path or just a file name
        file_path = Path(file_name)

        # Check if file_name is an absolute path, return this even if it does not exist
        if file_path.is_absolute():
            return file_path

        # Check if file_name contains any parent directories (i.e., is a relative path)
        # If so, return this even if it does not exist
        if file_path.parent != Path("."):  # noqa: PTH201
            # It's a relative path
            return (Path.cwd() / file_path).resolve()

        # Otherwise, assume it's just a file name and look for it in the current directory and the script directory
        current_dir = Path.cwd()
        file_path = current_dir / file_name
        if not file_path.exists():
            try:
                project_root_dir = SCCommon.get_project_root()
                file_path = project_root_dir / file_name
            except RuntimeError as e:
                error_msg = f"Cannot determine project root to locate file '{file_name}': {e}"
                raise RuntimeError(error_msg) from e
        return file_path

    @staticmethod
    def select_folder_location(folder_path: str | None = None, create_folder: bool = False) -> Path | None:  # noqa: FBT001, FBT002
        """Return an absolute folder path for the given (relative) folder path.

        If folder_path is None, return the project root folder.
        If folder_path is an absolute path, return it as a Path object.
        If folder_path is a relative path, return the absolute path based on the project root directory.

        Args:
            folder_path: The folder path to locate. Can be None, or a relative or absolute path.
            create_folder: If True, create the folder if it does not exist. Default is False.

        Raises:
            RuntimeError: If the project root cannot be determined or if folder creation fails.

        Returns:
            The full path to the folder as a Path object. None if folder_path is None and project root cannot be determined.
        """
        try:
            project_root = SCCommon.get_project_root()
        except RuntimeError as e:
            raise RuntimeError(e) from e

        if folder_path is None:
            return project_root

        selected_folder = Path(folder_path)

        # Check if folder_path is an absolute path, return this even if it does not exist
        if not selected_folder.is_absolute():
            selected_folder = (project_root / selected_folder).resolve()

        if create_folder and not selected_folder.exists():
            try:
                selected_folder.mkdir(parents=True, exist_ok=True)
            except OSError as e:
                error_msg = f"Error creating folder '{selected_folder}': {e}"
                raise RuntimeError(error_msg) from e

        return selected_folder

    @staticmethod
    def get_process_id() -> int:
        """Return the process ID of the current process.

        Returns:
            The process ID of the current process.
        """
        return os.getpid()



"""Unit tests for the sc_common module."""

import subprocess  # noqa: S404
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from sc_utility.sc_common import SCCommon


def test_is_valid_hostname_valid_ipv4():
    """Test is_valid_hostname with valid IPv4 addresses."""
    assert SCCommon.is_valid_hostname("192.168.1.1") is True
    assert SCCommon.is_valid_hostname("10.0.0.1") is True
    assert SCCommon.is_valid_hostname("127.0.0.1") is True
    assert SCCommon.is_valid_hostname("255.255.255.255") is True


def test_is_valid_hostname_invalid_ipv4():
    """Test is_valid_hostname with invalid IPv4 addresses."""
    assert SCCommon.is_valid_hostname("256.1.1.1") is False
    assert SCCommon.is_valid_hostname("192.168.1") is False
    assert SCCommon.is_valid_hostname("192.168.1.1.1") is False
    assert SCCommon.is_valid_hostname("192.168.1.a") is False


def test_is_valid_hostname_valid_ipv6():
    """Test is_valid_hostname with valid IPv6 addresses."""
    assert SCCommon.is_valid_hostname("::1") is True
    assert SCCommon.is_valid_hostname("2001:db8::1") is True
    assert SCCommon.is_valid_hostname("fe80::1") is True


def test_is_valid_hostname_valid_dns():
    """Test is_valid_hostname with valid DNS hostnames."""
    assert SCCommon.is_valid_hostname("example.com") is True
    assert SCCommon.is_valid_hostname("subdomain.example.com") is True
    assert SCCommon.is_valid_hostname("test-host.example.org") is True
    assert SCCommon.is_valid_hostname("localhost") is True
    assert SCCommon.is_valid_hostname("example.com.") is True  # FQDN


def test_is_valid_hostname_invalid_dns():
    """Test is_valid_hostname with invalid DNS hostnames."""
    assert SCCommon.is_valid_hostname("-example.com") is False
    assert SCCommon.is_valid_hostname("example-.com") is False
    assert SCCommon.is_valid_hostname("example..com") is False
    assert SCCommon.is_valid_hostname("ex@mple.com") is False
    assert SCCommon.is_valid_hostname("a" * 64 + ".com") is False  # Label too long


def test_is_valid_hostname_invalid_input():
    """Test is_valid_hostname with invalid input types."""
    assert SCCommon.is_valid_hostname("") is False
    assert SCCommon.is_valid_hostname(None) is False  # type: ignore[arg-type]
    assert SCCommon.is_valid_hostname(123) is False  # type: ignore[arg-type]
    assert SCCommon.is_valid_hostname([]) is False  # type: ignore[arg-type]


@patch("sc_utility.sc_common.subprocess.run")
@patch("sc_utility.sc_common.platform.system")
def test_ping_host_success_linux(mock_platform, mock_run):
    """Test ping_host success on Linux."""
    mock_platform.return_value = "Linux"
    mock_run.return_value = MagicMock(returncode=0)

    result = SCCommon.ping_host("192.168.1.1")

    assert result is True
    mock_run.assert_called_once_with(
        ["ping", "-c", "1", "-W", "1", "192.168.1.1"],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        shell=False,
        check=False
    )


@patch("sc_utility.sc_common.subprocess.run")
@patch("sc_utility.sc_common.platform.system")
def test_ping_host_success_windows(mock_platform, mock_run):
    """Test ping_host success on Windows."""
    mock_platform.return_value = "Windows"
    mock_run.return_value = MagicMock(returncode=0)

    result = SCCommon.ping_host("192.168.1.1")

    assert result is True
    mock_run.assert_called_once_with(
        ["ping", "-n", "1", "-W", "1", "192.168.1.1"],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        shell=False,
        check=False
    )


@patch("sc_utility.sc_common.subprocess.run")
@patch("sc_utility.sc_common.platform.system")
def test_ping_host_failure(mock_platform, mock_run):
    """Test ping_host failure."""
    mock_platform.return_value = "Linux"
    mock_run.return_value = MagicMock(returncode=1)

    result = SCCommon.ping_host("192.168.1.1")

    assert result is False


def test_ping_host_invalid_ip():
    """Test ping_host with invalid IP address."""
    with pytest.raises(RuntimeError, match="Invalid IP address"):
        SCCommon.ping_host("192.168.3.4.5")


def test_ping_host_untrusted_input():
    """Test ping_host with untrusted input."""
    with pytest.raises(RuntimeError, match="Invalid IP address"):
        SCCommon.ping_host("192.168.1.1; rm -rf /")


def test_check_internet_connection():
    """Test check_internet_connection."""
    assert SCCommon.check_internet_connection() is True  # This will depend on your actual internet connection


def test_is_probable_path_absolute_path():
    """Test is_probable_path with absolute paths."""
    assert SCCommon.is_probable_path("/usr/bin/python") is True
    assert SCCommon.is_probable_path("C:\\Windows\\System32") is True
    assert SCCommon.is_probable_path(Path("/usr/bin/python")) is True


def test_is_probable_path_relative_path():
    """Test is_probable_path with relative paths."""
    assert SCCommon.is_probable_path("./config.yaml") is True
    assert SCCommon.is_probable_path("../data/file.txt") is True
    assert SCCommon.is_probable_path("folder/file.txt") is True


def test_is_probable_path_file_extension():
    """Test is_probable_path with file extensions."""
    assert SCCommon.is_probable_path("config.yaml") is True
    assert SCCommon.is_probable_path("data.json") is True
    assert SCCommon.is_probable_path("script.py") is True


def test_is_probable_path_not_path():
    """Test is_probable_path with non-path strings."""
    assert SCCommon.is_probable_path("justtext") is False
    assert SCCommon.is_probable_path("no-extension") is False


@patch("sc_utility.sc_common.SCCommon.get_os")
@patch("sc_utility.sc_common.os.pathconf")
def test_is_probable_path_too_long(mock_pathconf, mock_get_os):
    """Test is_probable_path with path too long."""
    mock_get_os.return_value = "linux"
    mock_pathconf.return_value = 100

    long_path = "a" * 150 + ".txt"
    assert SCCommon.is_probable_path(long_path) is False


@patch("sc_utility.sc_common.Path.cwd")
def test_select_file_location_absolute_path(mock_cwd):  # noqa: ARG001
    """Test select_file_location with absolute path."""
    result = SCCommon.select_file_location("/etc/config.yaml")
    assert result == Path("/etc/config.yaml")


@patch("sc_utility.sc_common.Path.cwd")
def test_select_file_location_relative_path(mock_cwd):
    """Test select_file_location with relative path."""
    mock_cwd.return_value = Path("/home/user")

    result = SCCommon.select_file_location("config/settings.yaml")
    assert str(result).find("/home/user/config/settings.yaml")


@patch("sc_utility.sc_common.os.getpid")
def test_get_process_id(mock_getpid):
    """Test get_process_id."""
    mock_getpid.return_value = 12345
    assert SCCommon.get_process_id() == 12345
    mock_getpid.assert_called_once()


def test_get_process_id_integration():
    """Test get_process_id integration."""
    pid = SCCommon.get_process_id()
    assert isinstance(pid, int)
    assert pid > 0


# test_select_file_location_absolute_path()
# test_select_file_location_relative_path()



"""Validation schema for YAML configuration files."""

yaml_config_validation = {
    "Files": {
        "type": "dict",
        "schema": {
            "LogfileName": {"type": "string", "required": False, "nullable": True},
            "LogfileMaxLines": {"type": "number", "required": False, "nullable": True, "min": 0, "max": 100000},
            "TimestampFormat": {"type": "string", "required": False, "nullable": True},
            "LogProcessID": {"type": "boolean", "required": False, "nullable": True},
            "LogThreadID": {"type": "boolean", "required": False, "nullable": True},
            "LogfileVerbosity": {"type": "string", "required": True, "allowed": ["none", "error", "warning", "summary", "detailed", "debug", "all"]},
            "ConsoleVerbosity": {"type": "string", "required": True, "allowed": ["error", "warning", "summary", "detailed", "debug"]},
        },
    },
    "Email": {
        "type": "dict",
        "schema": {
            "EnableEmail": {"type": "boolean", "required": False, "nullable": True},
            "SendEmailsTo": {"type": "string", "required": False, "nullable": True},
            "SMTPServer":  {"type": "string", "required": False, "nullable": True},
            "SMTPPort": {"type": "number", "required": False, "nullable": True, "min": 25, "max": 10000},
            "SMTPUsername": {"type": "string", "required": False, "nullable": True},
            "SMTPPassword": {"type": "string", "required": False, "nullable": True},
            "SubjectPrefix": {"type": "string", "required": False, "nullable": True},
        },
    },
    "ShellyDevices": {
        "type": "dict",
        "schema": {
            "AllowDebugLogging": {"type": "boolean", "required": False, "nullable": True},
            "ResponseTimeout": {"type": "number", "required": False, "nullable": True, "min": 1, "max": 120},
            "RetryCount": {"type": "number", "required": False, "nullable": True, "min": 0, "max": 10},
            "RetryDelay": {"type": "number", "required": False, "nullable": True, "min": 1, "max": 10},
            "PingAllowed": {"type": "boolean", "required": False, "nullable": True},
            "SimulationFileFolder": {"type": "string", "required": False, "nullable": True},
            "WebhooksEnabled": {"type": "boolean", "required": False, "nullable": True},
            "WebhookHost": {"type": "string", "required": False, "nullable": True},
            "WebhookPort": {"type": "number", "required": False, "nullable": True},
            "WebhookPath": {"type": "string", "required": False, "nullable": True},
            "DefaultWebhooks": {
                "type": "dict",
                "required": False,
                "nullable": True,
                "schema": {
                    "Inputs": {
                        "type": "list",
                        "required": False,
                        "nullable": True,
                        "schema": {
                            "type": "string",
                            "required": True,
                        },
                    },
                    "Outputs": {
                        "type": "list",
                        "required": False,
                        "nullable": True,
                        "schema": {
                            "type": "string",
                            "required": True,
                        },
                    },
                    "Meters": {
                        "type": "list",
                        "required": False,
                        "nullable": True,
                        "schema": {
                            "type": "string",
                            "required": True,
                        },
                    },
                },
            },
            "Devices": {
                "type": "list",
                "required": True,
                "nullable": False,
                "schema": {
                    "type": "dict",
                    "schema": {
                        "Name": {"type": "string", "required": False, "nullable": True},
                        "Model": {"type": "string", "required": True},
                        "Hostname": {"type": "string", "required": False, "nullable": True},
                        "Port": {"type": "number", "required": False, "nullable": True},
                        "ID": {"type": "number", "required": False, "nullable": True},
                        "Simulate": {"type": "boolean", "required": False, "nullable": True},
                        "ExpectOffline": {"type": "boolean", "required": False, "nullable": True},
                        "Inputs": {
                            "type": "list",
                            "required": False,
                            "nullable": True,
                            "schema": {
                                "type": "dict",
                                "schema": {
                                    "Name": {"type": "string", "required": False, "nullable": True},
                                    "ID": {"type": "number", "required": False, "nullable": True},
                                    "Webhooks": {"type": "boolean", "required": False, "nullable": True},
                                },
                            },
                        },
                        "Outputs": {
                            "type": "list",
                            "required": False,
                            "nullable": True,
                            "schema": {
                                "type": "dict",
                                "schema": {
                                    "Name": {"type": "string", "required": False, "nullable": True},
                                    "ID": {"type": "number", "required": False, "nullable": True},
                                    "Webhooks": {"type": "boolean", "required": False, "nullable": True},
                                },
                            },
                        },
                        "Meters": {
                            "type": "list",
                            "required": False,
                            "nullable": True,
                            "schema": {
                                "type": "dict",
                                "schema": {
                                    "Name": {"type": "string", "required": False, "nullable": True},
                                    "ID": {"type": "number", "required": False, "nullable": True},
                                    "MockRate": {"type": "number", "required": False, "nullable": True},
                                },
                            },
                        },
                        "TempProbes": {
                            "type": "list",
                            "required": False,
                            "nullable": True,
                            "schema": {
                                "type": "dict",
                                "schema": {
                                    "Name": {"type": "string", "required": False, "nullable": True},
                                    "ID": {"type": "number", "required": False, "nullable": True},
                                    "RequiresOutput": {"type": "string", "required": False, "nullable": True},
                                },
                            },
                        },
                    },
                },
            },
        }
    }
}


"""Take any object and serialises it to a json file, converting data types as needed and adding hints to allow the original object to be recreated."""
import copy
import datetime as dt
import enum
import importlib
import json
import re
from pathlib import Path

from dateutil.parser import parse


class JSONEncoder:
    """Class to handle encoding and decoding of JSON data with special handling for datetime objects and enumerations."""
    @staticmethod
    def ready_dict_for_json(data: object) -> object:
        """Prepares a dict or list for JSON serialization.

        Args:
            data (object): The data to prepare.

        Raises:
            RuntimeError: If the data cannot be prepared.

        Returns:
            object: A dict or list ready for export to JSON.
        """
        def convert(obj):
            if isinstance(obj, dict):
                return {k: convert(v) for k, v in obj.items()}
            if isinstance(obj, list):
                return [convert(item) for item in obj]
            # Use the same logic as _encode_object for special types
            try:
                return JSONEncoder._encode_object(obj)
            except TypeError:
                return obj

        try:
            save_data = copy.deepcopy(data)
            save_data = JSONEncoder._add_datatype_hints(save_data)
        except (TypeError, ValueError) as e:
            raise RuntimeError from e
        return convert(save_data)

    @staticmethod
    def serialise_to_json(data) -> str:
        """Serialises the data to a JSON string, converting as needed.

        Args:
            data (object): The data to prepare.

        Raises:
            RuntimeError: If the data cannot be serialized.

        Returns:
            str: The JSON string representation of the data.
        """
        try:
            save_data = copy.deepcopy(data)
            save_data = JSONEncoder._add_datatype_hints(save_data)
            json_string = json.dumps(save_data, indent=4, default=JSONEncoder._encode_object)
        except (TypeError, ValueError) as e:
            raise RuntimeError from e
        else:
            return json_string

    @staticmethod
    def deserialise_from_json(json_string: str):
        """Deserialises the JSON string to an object, converting as needed.

        Args:
            json_string (str): The JSON string to deserialise.

        Raises:
            RuntimeError: If the data cannot be deserialized.

        Returns:
            return_data(object): The deserialized object.
        """
        try:
            json_data = json.loads(json_string)
            return_data = JSONEncoder.decode_object(json_data)
        except (json.JSONDecodeError, ValueError) as e:
            raise RuntimeError from e
        else:
            return return_data

    @staticmethod
    def save_to_file(data, file_path: Path) -> bool:
        """Saves the date to a JSON file, converting as needed.

        Args:
            data (object): The data to prepare.
            file_path (Path): The path to the JSON file to be created.

        Raises:
            RuntimeError: If the data cannot be serialized.

        Returns:
            result (bool): True if the pricing data was saved, False if not.
        """
        try:
            temporary_path = file_path.with_suffix(".tmp")

            with temporary_path.open("w", encoding="utf-8") as json_file:
                save_data = copy.deepcopy(data)
                save_data = JSONEncoder._add_datatype_hints(save_data)
                json.dump(save_data, json_file, indent=4, default=JSONEncoder._encode_object)

            temporary_path.replace(file_path)
        except (TypeError, ValueError, OSError) as e:
            raise RuntimeError from e
        return True

    @staticmethod
    def read_from_file(file_path: Path) -> object | None:
        """Reads the JSON data from a file and decodes it.

        Args:
            file_path (Path): The path to the JSON file.

        Raises:
            RuntimeError: If the data cannot be read or decoded.

        Returns:
            dict: The decoded JSON data or None if the file does not exist.
        """
        if not file_path.exists():
            return None

        try:
            with file_path.open("r", encoding="utf-8") as json_file:
                json_data = json.load(json_file)
                return_data = JSONEncoder.decode_object(json_data)
                return return_data
        except (json.JSONDecodeError, OSError) as e:
            raise RuntimeError from e

    @staticmethod
    def _add_datatype_hints(obj):
        """Add datetime hints to the object before it's serialized.

        Args:
            obj: The object to convert.

        Returns:
            The object with hints added
        """
        if isinstance(obj, dict):
            # Build a new dict so we can place hint-keys immediately after their associated key
            new_obj: dict = {}
            for k, v in obj.items():
                # Recurse for nested containers first
                if isinstance(v, (dict, list)):
                    v_conv = JSONEncoder._add_datatype_hints(v)
                else:
                    v_conv = v

                # Add the original key/value
                new_obj[k] = v_conv

                # Immediately add any hint key right after
                if isinstance(v_conv, (dt.date, dt.datetime, dt.time)):
                    new_obj[f"{k}__datatype"] = type(v_conv).__name__  # e.g. "date", "datetime", "time"
                elif isinstance(v_conv, enum.Enum):
                    enum_cls = v_conv.__class__
                    new_obj[f"{k}__enum"] = f"{enum_cls.__module__}.{enum_cls.__name__}.{v_conv.name}"

            return new_obj
        if isinstance(obj, list):
            return [JSONEncoder._add_datatype_hints(item) for item in obj]
        return obj

    @staticmethod
    def _encode_object(obj):
        """Convert the object to JSON serialisable format if it's a date or datetime. This function is use by json.dump().

        Args:
            obj: The object to convert.

        Raises:
            TypeError: If the object is not serializable.

        Returns:
            The JSON serializable representation of the object.
        """
        if isinstance(obj, (dt.datetime, dt.date, dt.time)):
            return obj.isoformat()
        if isinstance(obj, enum.Enum):
            # Store enum by its value; hint added in _add_datatype_hints enables reconstruction
            return obj.value
        error_msg = f"Type {type(obj)} not serializable"
        raise TypeError(error_msg)

    @staticmethod
    def decode_object(obj):  # noqa: PLR0912, PLR0915
        """Convert the object back to its original form, including date and datetime objects.

        Args:
            obj (obj): The object (list, dict, etc.) to convert.

        Returns:
            object (obj): The original object.
        """
        if isinstance(obj, dict):  # noqa: PLR1702
            for k, v in list(obj.items()):
                if isinstance(v, str):
                    # See if there's a enum hint for this key
                    enum_hint_key = f"{k}__enum"
                    enum_hint = obj.get(enum_hint_key)
                    obj.pop(enum_hint_key, None)  # remove the enum hint from the object
                    if enum_hint:
                        try:
                            if "." in enum_hint and enum_hint.count(".") >= 2:
                                # New format: module.class.key_name
                                parts = enum_hint.rsplit(".", 2)
                                module_name, class_name, key_name = parts
                                enum_module = importlib.import_module(module_name)
                                enum_cls = getattr(enum_module, class_name)
                                if issubclass(enum_cls, enum.Enum):
                                    obj[k] = getattr(enum_cls, key_name)
                                    continue
                            else:
                                # Old format: module.class (fallback)
                                module_name, class_name = enum_hint.rsplit(".", 1)
                                enum_module = importlib.import_module(module_name)
                                enum_cls = getattr(enum_module, class_name)
                                if issubclass(enum_cls, enum.Enum):
                                    obj[k] = enum_cls(v)
                                    continue
                        except Exception:  # noqa: BLE001, S110
                            # If reconstruction fails, leave as-is and fall through
                            pass

                    # See if there's a datatype hint for this key
                    datatype_hint_key = f"{k}__datatype"
                    datatype_hint = obj.get(datatype_hint_key)
                    obj.pop(datatype_hint_key, None)  # remove the datatype hint from the object
                    if datatype_hint == "date":
                        try:
                            obj[k] = dt.date.fromisoformat(v)
                            continue
                        except ValueError:
                            pass
                    elif datatype_hint == "datetime":
                        try:
                            obj[k] = dt.datetime.fromisoformat(v)
                            continue
                        except ValueError:
                            pass
                    elif datatype_hint == "time":
                        try:
                            obj[k] = dt.time.fromisoformat(v)
                            continue
                        except ValueError:
                            pass

                    # If the string is a date-only value like "YYYY-MM-DD", decode to a date
                    if re.match(r"^\d{4}-\d{2}-\d{2}$", v):
                        try:
                            obj[k] = dt.date.fromisoformat(v)
                            continue
                        except ValueError:
                            # fall through to the full parse attempt
                            pass

                    # Only try to parse if the string contains date/time-like patterns
                    # This prevents parsing strings like "Sat,Sun,Tue" as dates
                    if re.search(r"\d", v) and re.search(r"[:/\-]", v):
                        try:
                            dt_obj = parse(v, fuzzy=False)
                            # If time part is zero, treat as date
                            if dt_obj.time() == dt.time(0, 0):
                                obj[k] = dt_obj.date()
                            else:
                                obj[k] = dt_obj
                        except (ValueError, TypeError):
                            pass    # Just ignore
                elif isinstance(v, (dict, list)):
                    obj[k] = JSONEncoder.decode_object(v)
        elif isinstance(obj, list):
            return [JSONEncoder.decode_object(item) for item in obj]
        return obj



"""
Spello Consulting Logging Module.

Provides general purpose logging functions.
"""
import datetime as dt
import inspect
import smtplib
import ssl
import sys
import threading
import traceback
from dataclasses import dataclass
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from pathlib import Path

from sc_utility.sc_common import SCCommon
from sc_utility.sc_date_helper import DateHelper


@dataclass
class NotifiableIssue:
    entity: str
    issue_type: str
    first_reported: dt.datetime
    last_reported: dt.datetime
    send_delay: int
    message: str
    email_sent: bool = False


class SCLogger:
    """A class to handle logging messages with different verbosity levels."""

    def __init__(self, logger_settings: dict):
        """
        Initializes the logger with configuration settings.

        The structure of logger_settings is as follows:
        {
            "logfile_name": Optional[str],  # The name of the log file. If None, no file logging will be done.
            "file_verbosity": Optional[str],  # Verbosity level for file logging
            "console_verbosity": Optional[str],  # Verbosity level for console logging
            "max_lines": Optional[int],  # Maximum number of lines to keep in the log file
            "log_process_id": Optional[bool],  # If True, include the process ID in log messages. Defaults to False.
            "log_thread_id": Optional[bool],  # If True, include the thread ID in log messages. Defaults to False.
        }

        Args:
            logger_settings (Optional[dict]): A dictionary containing logger settings
        """
        # Make a note of the app directory
        self.app_dir = self.client_dir = SCCommon.get_project_root()

        # Setup the path to the fatal error tracking file
        self.fatal_error_file_path = self.app_dir / f"{self.app_dir.name}_fatal_error.txt"

        # Use the register_email_settings method to set up email settings
        self.email_settings = None

        self.verbosity_levels = {
            "none": 0,
            "error": 1,
            "warning": 2,
            "summary": 3,
            "detailed": 4,
            "debug": 5,
            "all": 6,
        }
        # Attributes set in the initialise_logger method
        self.logfile_name = None
        self.logfile_path = None
        self.file_verbosity = "summary"
        self.console_verbosity = "summary"
        self.max_lines = 1000
        self.timestamp_format = "%Y-%m-%d %H:%M:%S"
        self.log_process_id = False
        self.log_thread_id = False
        self.file_logging_enabled = False
        self.notifiable_issues: list[NotifiableIssue] = []

        self.initialise_settings(logger_settings)

    def initialise_settings(self, logger_settings: dict):
        """
        Set / reset the logger configuration settings. See the __init__() method for the structure of logger_settings.

        Args:
            logger_settings (Optional[dict]): A dictionary containing logger settings.
        """
        self.logfile_name = logger_settings.get("logfile_name")
        self.file_verbosity = logger_settings.get("file_verbosity", "summary")
        self.console_verbosity = logger_settings.get("console_verbosity", "summary")
        self.max_lines = logger_settings.get("max_lines", 1000)
        self.timestamp_format = logger_settings.get("timestamp_format", "%Y-%m-%d %H:%M:%S")
        self.log_process_id = logger_settings.get("log_process_id", False)
        self.log_thread_id = logger_settings.get("log_thread_id", False)

        # See if logfile writing is required
        self.file_logging_enabled = self.logfile_name is not None

        if self.file_logging_enabled and self.logfile_name is not None:
            self.logfile_path = SCCommon.select_file_location(self.logfile_name)

            # Truncate the log file if it exists
            self._initialise_monitoring_logfile()

    def _initialise_monitoring_logfile(self) -> None:
        """Initialise the monitoring log file. If it exists, truncate it to the max number of lines."""
        if not self.file_logging_enabled:
            return

        self.trim_logfile()

    def trim_logfile(self) -> None:
        """Trims the log file to the maximum number of lines specified."""
        if not self.file_logging_enabled:
            return

        assert self.logfile_path is not None, "Log file path must be set before trimming the log file."
        if self.logfile_path.exists() and self.max_lines is not None and self.max_lines > 0:
            # Monitoring log file exists - truncate excess lines if needed.
            with self.logfile_path.open(encoding="utf-8") as file:
                lines = file.readlines()

                if len(lines) > self.max_lines:
                    # Keep the last max_lines rows
                    keep_lines = lines[-self.max_lines:] if len(lines) > self.max_lines else lines

                    # Overwrite the file with only the last 1000 lines
                    with self.logfile_path.open("w", encoding="utf-8") as file2:
                        file2.writelines(keep_lines)

    def log_message(self, message: str, verbosity: str = "summary") -> None:
        """Writes a log message to the console and/or a file based on verbosity settings.

        Args:
            message (str): The message to log.
            verbosity (str): The verbosity level for the message. Must be one of "none", "error", "warning", "summary", "detailed", "debug", or "all".

        Raises:
            ValueError: If an invalid verbosity level is provided.

        """
        if verbosity not in self.verbosity_levels:
            exception_msg = f"log_message(): Invalid verbosity passed, must be one of {list(self.verbosity_levels.keys())}."
            raise ValueError(exception_msg)

        file_verbosity = self.file_verbosity if self.file_verbosity is not None else "none"
        console_verbosity = self.console_verbosity if self.console_verbosity is not None else "none"
        logfile_level = self.verbosity_levels.get(file_verbosity, 0)
        console_level = self.verbosity_levels.get(console_verbosity, 0)
        message_level = self.verbosity_levels.get(verbosity, 0)

        process_str = ""
        if self.log_process_id and self.log_thread_id:
            process_str = f"[Proc {SCCommon.get_process_id()}, Thread {threading.current_thread().name}] "
        elif self.log_process_id:
            process_str = f"[Proc {SCCommon.get_process_id()}] "
        elif self.log_thread_id:
            process_str = f"[Thread {threading.current_thread().name}] "

        # Deal with console message first
        if console_level >= message_level and console_level > 0:
            if verbosity == "error":
                print(f"ERROR: {process_str}{message}", file=sys.stderr)
            elif verbosity == "warning":
                print(f"WARNING: {process_str}{message}")
            else:
                print(f"{process_str}{message}")

        # Now write to the log file if needed
        if self.file_logging_enabled:
            assert self.logfile_path is not None, "Log file path must be set before trimming the log file."
            error_str = " ERROR" if verbosity == "error" else " WARNING" if verbosity == "warning" else ""
            if logfile_level >= message_level and logfile_level > 0:
                with self.logfile_path.open("a", encoding="utf-8") as file:
                    if not message:
                        file.write("\n")
                    else:
                        # Use the local timezone for the log timestamp
                        file.write(f"{DateHelper.now().strftime(self.timestamp_format)}{error_str}: {process_str}{message}\n")

    def register_email_settings(self, email_settings: dict | None) -> None:
        """
        Registers email settings for sending emails.

        Use the SCConfigManager.get_email_settings() method to get a dictionary containing the email settings. Otherwise, you can pass a
        dictionary directly to this method with these keys:
            SendEmailsTo: str, the email address to send emails to
            SMTPServer: str, the SMTP server address
            SMTPPort: int, the SMTP server port (optional, defaults to 587)
            SMTPUsername: str, the username for the SMTP server
            SMTPPassword: str, the password for the SMTP server (preferably an App Password)
            SubjectPrefix: str, a prefix for email subjects (optional, default to None)

        Args:
            email_settings (Optional[dict], optional): A dictionary containing email settings. If empty or None, no email settings will be registered.

        Raises:
            TypeError: If email_settings is not a dictionary.
            ValueError: If any required keys are missing from the email_settings dictionary.

        """
        if email_settings == {} or email_settings is None:
            return  # No email settings provided, so skip registration

        if not isinstance(email_settings, dict):
            msg = "register_email_settings(): Email settings must be a dictionary."
            raise TypeError(msg)

        # Check if all required keys are present
        required_keys = ["SendEmailsTo", "SMTPServer", "SMTPUsername", "SMTPPassword"]
        for key in required_keys:
            if key not in email_settings:
                msg = f"register_email_settings(): Missing required email setting: {key}"
                raise ValueError(msg)

        # Store the email settings in the config object
        self.email_settings = email_settings

    def send_email(self, subject: str, body: str | Path, test_mode: bool = False) -> bool:  # noqa: FBT001, FBT002, PLR0911, PLR0912, PLR0915
        """
        Sends an email using the SMTP server previously specified in register_email_settings().

        Args:
            subject (str): The subject of the email.
            body (str): The body of the email. This argument can be one of 4 things:
                1. A string containing the HTML body of the email
                2. A string or Path containing the path to an HTML file to read the body from
                3. A string containing the text body of the email
                4. A string or Path containing the path to an text file to read the body from
            test_mode (bool, optional): If True, the email will not be sent, but a message will be logged indicating that the email would have been sent. Defaults to False.

        Returns:
            result (bool): True if the email was sent successfully, False otherwise.
        """
        if self.email_settings is None:
            return False  # No email settings registered, so skip sending the email

        # First confirm that the body argument was passed as a string
        if not isinstance(body, str) and not isinstance(body, Path):
            self.log_fatal_error("body argument must be a string containing the body content or a file path.")
            return False

        # See if the body string is something that looks like a file path
        try:
            payload_path = SCCommon.select_file_location(str(body))
        except OSError:
            # Nothing to do here
            payload_path = None

        payload = body  # Default to the body as text content
        payload_type = "plain"

        # Default to treating the body a file path and see if we can resolve it
        if payload_path is not None:
            if payload_path.exists():
                # If the body is a file path, read the content
                with payload_path.open("r", encoding="utf-8") as file:
                    payload = file.read()

                # Determine the payload type based on the file extension
                if payload_path.suffix.lower() == ".html":
                    payload_type = "html"
                elif payload_path.suffix.lower() == ".txt":
                    payload_type = "plain"
                else:
                    self.log_fatal_error(f"Unsupported file type for email body: {payload_path.suffix}")
                    return False
            else:
                payload_path = None

        if payload_path is None:
            # If fall through to here, then the body was not a file path or the file did not exist
            # Assume that the body is a string containing the content
            payload = str(body)  # Default to the body as text content
            if payload.startswith(("<html", "<!DOCTYPE html")):
                payload_type = "html"
            else:
                payload_type = "plain"

        # Load the Gmail SMTP server configuration
        sender_email = self.email_settings.get("SMTPUsername")
        send_to = self.email_settings.get("SendEmailsTo")
        smtp_server = self.email_settings.get("SMTPServer")
        smtp_port = self.email_settings.get("SMTPPort", 587)
        smtp_password = self.email_settings.get("SMTPPassword")

        if not sender_email or not send_to or not smtp_server or not smtp_port or not smtp_password:
            self.log_fatal_error("send_email(): Email settings are incomplete. 'SMTPUsername', 'SendEmailsTo', 'SMTPServer' and 'SMTPPassword' must be provided.")
            return False

        try:
            # Create the email
            msg = MIMEMultipart()
            msg["From"] = sender_email
            msg["To"] = send_to
            if self.email_settings.get("SubjectPrefix", None) is not None:
                msg["Subject"] = self.email_settings.get("SubjectPrefix", "") + subject
            else:
                msg["Subject"] = subject

            # Attach the payload as either plain text or HTML
            part = MIMEText(payload, payload_type)  # type: ignore[arg-type]
            msg.attach(part)

            # Connect to the Gmail SMTP server
            with smtplib.SMTP(host=smtp_server, port=smtp_port, timeout=10) as server:
                server.starttls()  # Upgrade the connection to secure
                server.login(sender_email, smtp_password)  # Log in using App Password
                if test_mode:
                    self.log_message(f"Test mode: Email with subject '{msg['Subject']}' would be sent to {send_to}.", "summary")
                else:
                    server.sendmail(sender_email, send_to, msg.as_string())  # Send the email

        except TimeoutError as e:
            self.log_message(f"send_email(): Timeout sending email: {e}", "error")
            return False
        except smtplib.SMTPAuthenticationError as e:
            self.log_message(f"send_email(): Authentication failed: {e}", "error")
            return False
        except smtplib.SMTPRecipientsRefused as e:
            self.log_message(f"send_email(): Recipients refused: {e}", "error")
            return False
        except smtplib.SMTPSenderRefused as e:
            self.log_message(f"send_email(): Sender refused: {e}", "error")
            return False
        except (smtplib.SMTPException, ssl.SSLError) as e:
            self.log_message(f"send_email(): SMTP/SSL error: {e}", "error")
            return False
        except (OSError, ConnectionRefusedError, smtplib.SMTPServerDisconnected) as e:
            self.log_message(f"send_email(): Network error sending email: {e}", "error")
            return False
        except Exception as e:  # noqa: BLE001
            self.log_message(f"send_email(): Unexpected error: {e}", "error")
            return False
        else:
            return True  # Email sent successfully

    def log_fatal_error(self, message: str, report_stack: bool = False, calling_function: str | None = None) -> None:  # noqa: FBT001, FBT002
        """
        Log a fatal error, send an email if configured to so and then exit the program.

        Args:
            message (str): The error message to log.
            report_stack (Optional[bool], optional): If True, include the stack trace in the log message.
            calling_function (Optional[str], optional): The name of the function that called this method, if known. If None, the calling function will be determined automatically.

        Raises:
            SystemExit: Exits the program with a status code of 1 after logging the fatal error.

        """  # noqa: DOC502
        function_name = None
        if calling_function is None:
            stack = inspect.stack()
            # Get the frame of the calling function
            calling_frame = stack[1]
            # Get the function name
            function_name = calling_frame.function
            if function_name == "<module>":
                function_name = "main"
            # Get the class name (if it exists)
            class_name = None
            if "self" in calling_frame.frame.f_locals:
                class_name = calling_frame.frame.f_locals["self"].__class__.__name__
                full_reference = f"{class_name}.{function_name}()"
            else:
                full_reference = function_name + "()"
        else:
            full_reference = calling_function + "()"

        if report_stack:
            stack_trace = traceback.format_exc()
            message += f"\n\nStack trace:\n{stack_trace}"

        self.log_message(f"Function {full_reference}: FATAL ERROR: {message}", "error")

        # Try to send an email if configured to do so and we haven't already sent one for a fatal error
        # Don't send concurrent error emails
        if function_name != "send_email" and not self.get_fatal_error():
            self.send_email(
                f"{self.app_dir.name} terminated with a fatal error",
                f"{message} \nAdditional emails will not be sent for concurrent errors.",
            )

        # record the error in in a file so that we keep track of this next time round
        self.set_fatal_error(message)

        # Exit the program
        sys.exit(1)

    def get_fatal_error(self) -> bool:
        """Returns True if a fatal error was previously reported, false otherwise."""
        return Path(self.fatal_error_file_path).exists()

    def clear_fatal_error(self) -> bool:
        """
        Clear a previously logged fatal error.

        Returns:
            result (bool): True if the file was deleted, False if it did not exist.
        """
        if Path(self.fatal_error_file_path).exists():
            Path(self.fatal_error_file_path).unlink()
            return True
        return False

    def set_fatal_error(self, message: str) -> None:
        """
        Create a fatal error tracking file and write the message to it.

        Args:
            message (str): The error message to write to the fatal error file.
        """
        with Path(self.fatal_error_file_path).open("w", encoding="utf-8") as file:
            file.write(message)

    def report_notifiable_issue(self, entity: str, issue_type: str, send_delay: int, message: str) -> bool:
        """
        Log a notifiable issue and send an email if configured to do so.

        When a new combination of entity and issue_type is reported, it will be noted including the current timestamp.
        If one or more issues of the same type are reported for the same entity at least send_delay seconds after the last notification,
        an email will be sent including the first time this issue was reported. Subsequent issues of the same type for the same
        entity will be ignored until the issue is cleared.

        Args:
            entity (str): The entity reporting the issue (e.g., "Device 123", "Module", etc.)
            issue_type (str): A brief description of the issue (e.g. "Offline").
            send_delay (int): The delay in seconds before sending the email
            message (str): The issue message to log and email.

        Returns:
            result (bool): True if the email was sent, False otherwise.
        """
        # Search self.notifiable_issues to see if we have a match by entity and issue_type
        now = DateHelper.now()
        issue_found = False
        for issue in self.notifiable_issues:
            if issue.entity == entity and issue.issue_type == issue_type:
                issue_found = True
                # We have a match - see if we need to send an email
                if (now - issue.first_reported).total_seconds() >= issue.send_delay:
                    # Update the last reported time
                    issue.last_reported = now

                    # Send the email if not already sent
                    if not issue.email_sent:
                        email_subject = f"{self.app_dir.name} - {entity} {issue_type} issue"
                        email_body = f"{entity}: {message}\n\nFirst reported: {issue.first_reported.strftime('%Y-%m-%d %H:%M')}"
                        if self.send_email(email_subject, email_body):
                            issue.email_sent = True
                            return True

                break

        if not issue_found:
            # No existing issue found, create a new one
            new_issue = NotifiableIssue(
                entity=entity,
                issue_type=issue_type,
                first_reported=now,
                last_reported=now,
                send_delay=send_delay,
                message=message,
            )
            self.notifiable_issues.append(new_issue)

        return False

    def clear_notifiable_issue(self, entity: str, issue_type: str) -> bool:
        """
        Clear a notifiable issue.

        If an email was previously sent for this issue, this function will send a "cleared" notification email.

        Args:
            entity (str): The entity reporting the issue (e.g., "Device 123", "Module", etc.)
            issue_type (str): A brief description of the issue (e.g. "Offline").

        Returns:
            result (bool): True if the issue was found and cleared, False otherwise.
        """
        # Search self.notifiable_issues to see if we have a match by entity and issue_type
        for issue in self.notifiable_issues:
            if issue.entity == entity and issue.issue_type == issue_type:
                # Send a cleared email if one was previously sent
                if issue.email_sent:
                    email_subject = f"{self.app_dir.name} - {entity} {issue_type} issue resolved"
                    email_body = f"{entity}: The {issue_type} issue reported on {issue.first_reported.strftime('%Y-%m-%d %H:%M')} has been resolved."
                    self.send_email(email_subject, email_body)

                self.notifiable_issues.remove(issue)
                return True

        return False
